{-----------------------------------------------------------------------------}
{ Podprogram za urejanje z algoritmom quicksort                               }
{-----------------------------------------------------------------------------}

{ Ta funkcija naj primerja elementa 'E1' in 'E2' in:
  - vrne -1, ce je E1 < E2;
  - vrne +1, ce je E1 > E2;
  - vrne 0, ce je E1 = E2. }
type TCompareFunc = function(var E1, E2): Integer;

{ Naslednji podprogram uredi tabelo 'Data'.  V njej mora biti 'NumElements'
  elementov, vsak od njih pa mora biti dolg 'Width' bajtov.  'Compare' naj 
  kaze na funkcijo, ki zna primerjati te elemente. }
procedure QuickSort(var Data; NumElements, Width: Integer; Compare: TCompareFunc);
type
	PArr = ^TArr;
	TArr = packed array[0..0] of Byte;
var
	Temp, M: PArr;

	procedure Qs(Data: PArr; L, R, Width: Integer; Compare: TCompareFunc);
	var
		I, J: Integer;
		A: Integer;
	begin
		while R - L > 1 do
			begin
			{ Partition }
			I := L; J := R; Move(Data^[((L + R) div 2) * Width], M^, Width);
			while I < J do
				begin
				while Compare(Data^[I * Width], M^) < 0 do Inc(I);
				while Compare(Data^[J * Width], M^) > 0 do Dec(J);
				if I < J then 
					begin 
					Move(Data^[I * Width], Temp^, Width);
					Move(Data^[J * Width], Data^[I * Width], Width);
					Move(Temp^, Data^[J * Width], Width); Inc(I); Dec(J); 
					end;
				end;
			{ Recursive calls }
			if J - L > R - I then begin Qs(Data, L, J, Width, Compare); L := I; end
			else begin Qs(Data, I, R, Width, Compare); R := J; end;
			end;
		{ Insertion sort }
		for I := L + 1 to R do 
			begin
			J := I - 1; Move(Data^[I * Width], Temp^, Width);
			while J >= L do 
				begin
				if Compare(Data^[J * Width], Temp^) <= 0 then Break;
				Move(Data^[J * Width], Data^[(J + 1) * Width], Width); Dec(J); 
				end;
			{ WriteLn(I, ' -> ', J + 1); }
			Move(Temp^, Data^[(J + 1) * Width], Width); 
			end;
	end;

begin
	GetMem(Temp, Width); GetMem(M, Width);
	Qs(PArr(@Data), 0, NumElements - 1, Width, Compare);
	FreeMem(Temp, Width); FreeMem(M, Width);
end;

{-----------------------------------------------------------------------------}
{ Primer uporabe podprograma QuickSort                                        }
{-----------------------------------------------------------------------------}

{ Funkcija za primerjanje nizov, ki jo bomo podtaknili QuickSortu. }
function MyStrCompareFunc(var E1, E2): Integer;
type PString = ^String;
var S1, S2: PString;
begin
	S1 := PString(@E1); S2 := PString(@E2);
	{ WriteLn(S1^, ' ', S2^); }
	if S1^ < S2^ then MyStrCompareFunc := -1
	else if S1^ > S2^ then MyStrCompareFunc := 1
	else MyStrCompareFunc := 0;
end;

{ Primer 1: uredimo tabelo desetih nizov. }
{
var
	MyArr: packed array[1..10] of String; I: Integer;
begin
	MyArr[1] := 'ena';
	MyArr[2] := 'dve';
	MyArr[3] := 'tri';
	MyArr[4] := 'stiri';
	MyArr[5] := 'pet';
	MyArr[6] := 'sest';
	MyArr[7] := 'sedem';
	MyArr[8] := 'osem';
	MyArr[9] := 'devet';
	MyArr[10] := 'deset';
	QuickSort(MyArr, 10, sizeof(String), MyStrCompareFunc);
	for I := 1 to 10 do WriteLn(MyArr[I]);
end.
}

{ Primer 2: zgeneriramo nekaj nakljucnih nizov in jih uredimo. }
const Count = 10000;
var
	MyArr: packed array[1..Count] of String; I, J: Integer;
begin
	for I := 1 to Count do
	begin
		MyArr[I] := '...';
		for J := 1 to 3 do
			MyArr[I][J] := Char(Integer('a') + Random(26));
		WriteLn(MyArr[I]);
	end;
	WriteLn('--');
	QuickSort(MyArr, Count, sizeof(String), MyStrCompareFunc);
	for I := 1 to Count do WriteLn(MyArr[I]);
end.
